import '../utils/tools.dart';
import '../base/token.dart';
import '../base/parser.dart';
import '../base/lexer.dart';
import '../base/event.dart';
import '../base/types.dart';
import '../base/unprocessed_token.dart';

class Position {
  int from;
  Position(this.from);
}

typedef RexMatch = RegExpMatch Function(String text, int pos);
const CONFIG_EVENT_DISPATCHER = 'eventDispatcher';
const CONFIG_USECACHE = 'useCache';
const CONFIG_SAVING_RUNTIME_CONTEXT = 'savingRuntimeContext';
const CONFIG_STATE_WILL_LIST_TOKENS = 'stateWillListTokens';
const CONFIG_DEBUGGABLE = 'debuggable';
const CONFIG_ENUMSTRICT = 'enumStrict';
const CONFIG_GROUPCOUNT_STRICTCHECK = 'groupCountStrictCheck';
const CONFIG_MATCHLENGTH_STRICTCHECK = 'matchLengthStrictCheck';
const CONFIG_DEBUGPRINT_THROWERROR = 'debugPrintThrowError';

abstract class RegexLexer<T extends Parse> extends Lexer {
  RegexLexer({
    stripnl = true,
    stripall = false,
    ensurenl = true,
    tabsize = 0,
    encoding = 'guess',
    debuggable = false,
  }) : super(
            stripnl: stripnl,
            stripall: stripall,
            ensurenl: ensurenl,
            tabsize: tabsize,
            encoding: encoding,
            debuggable: debuggable);
  RegExpFlags get flags;
  Map<String, List<T>> get parses;
  Map<String, dynamic>? config;

  /// runtimeContext is the deepest context the parser can go.
  /// each parser has two part of parsing rules. Single parses and Common parses.
  /// if you nested lexer with another lexer use [JParse.lexer], the context go deeper.
  /// Each parser's single parses is isolated but common parses will merge.
  /// So the deeper goes the common parse becomes larger.That's what we want when searching
  /// a certain state.
  /// However the single part belongs to each parser. and the state name maybe conflict.
  /// but that doesn't matter, Because when a parser is running,it automaticlly ignored other parsers' own defination.
  /// When you start parsing from a LexerMain there's only one root Lexer and many LibrayLexers ,don't worry you'll lose any context.
  Map<String, List<T>>? _runtimeContext;

  /// a cacheContext is only for those case expanding the same parse package.
  /// the context stays inside of this parser and this will not cause conflict.
  /// different parser has its own cacheContext, they should not be merged anytime for the risk of naming conflict.
  final Map<String, List<T>> _cacheContext = {};
  bool get supportEventDispatching =>
      config?.containsKey(CONFIG_EVENT_DISPATCHER) ?? false;
  RawEventDispatcher? get configEventDispatcher =>
      supportEventDispatching ? config![CONFIG_EVENT_DISPATCHER] : null;
  bool get useCache => config?.containsKey(CONFIG_USECACHE) ?? false
      ? (config![CONFIG_USECACHE] as bool)
      : true;
  bool get configSaveRuntimeContext =>
      config?.containsKey(CONFIG_SAVING_RUNTIME_CONTEXT) ?? false
          ? config![CONFIG_SAVING_RUNTIME_CONTEXT]
          : true;
  List<dynamic>? get configListTokenStateNames =>
      config?.containsKey(CONFIG_STATE_WILL_LIST_TOKENS) ?? false
          ? config![CONFIG_STATE_WILL_LIST_TOKENS] as List
          : null;
  bool get configDebuggable => config?.containsKey(CONFIG_DEBUGGABLE) ?? false
      ? config![CONFIG_DEBUGGABLE]
      : debuggable;
  bool get configEnumStrict => config?.containsKey(CONFIG_ENUMSTRICT) ?? false
      ? config![CONFIG_ENUMSTRICT]
      : false;
  bool get configGroupCountStrictCheck =>
      config?.containsKey(CONFIG_GROUPCOUNT_STRICTCHECK) ?? false
          ? config![CONFIG_GROUPCOUNT_STRICTCHECK]
          : true;
  bool get configMatchLengthStrictCheck =>
      config?.containsKey(CONFIG_MATCHLENGTH_STRICTCHECK) ?? false
          ? config![CONFIG_MATCHLENGTH_STRICTCHECK]
          : true;
  bool get configDebugPrintThrowError =>
      config?.containsKey(CONFIG_DEBUGPRINT_THROWERROR) ?? false
          ? config![CONFIG_DEBUGPRINT_THROWERROR]
          : false;

  void printThrow(String text) {
    if (configDebugPrintThrowError) throw text;
    print(text);
  }

  /// You don't have to merge like this
  /// return { ...currentCommon, [your Def]: bla bla bla}
  /// just give your own lib common parse rules
  /// => { [your Def]: bla bla bla }
  /// currentCommon is for you to check what have been loaded if Needed.
  Map<String, List<T>> commonparses(Map<String, List<T>>? currentCommon);

  List<String> _enumAllGiven(DynamicToken? enumToken, String stateName) {
    // assert(enumToken.isEnum == true);
    if (enumToken == null) return [];
    if (enumToken.isEnum) {
      if (enumToken.enums!.isNotEmpty) {
        return enumToken.enums!;
      } else {
        var exist = <String>[];
        _tokensUnprocessed?.forEach((element) {
          if (
              //去重
              !exist.contains(element.match) &&
                  enumToken.enumsEqual(element.token) &&
                  element.stateName == stateName) {
            exist.add(element.match);
          }
        });
        return exist;
      }
    } else {
      var exist = <String>[];
      return _tokensUnprocessed
              ?.where((element) {
                var mark = element.stateName + element.match;
                if (
                    //去重
                    !exist.contains(mark) &&
                        //相等
                        enumToken == element.token &&
                        //严格模式比对所在 state 名称
                        (configEnumStrict
                            ? element.stateName == stateName
                            : true)) {
                  return true;
                }
                return false;
              })
              .map((item) => item.match)
              .toList() ??
          [];
    }

    // if(configEnumStrict == false){
    //   return _tokensUnprocessed.where((element) => enumToken == element.token)
    // }else return [];
  }

  List<String> splitAutoCompleting(List<String> splitText, String stateName,
      {int matchMode = 3}) {
    /// step1: find matches;
    var matches =
        splitAutoCompletingMatches(splitText, stateName, matchMode: matchMode);

    /// step2: sort
    /// 从左到右 -> 优先
    /// start 小 ->
    /// input.length 小 ->
    matches.sort((left, right) {
      if (left[0].input.length < right[0].input.length) {
        return -1;
      }
      return 1;
    });

    /// step3: 用 Set 去重
    return (<String>{}..addAll(matches
            .map<String>((m) => m[0].input)
            .where((element) => element != splitText.join())))
        .toList();
  }

  /// Match wd10 to width_10
  /// matchMode:
  /// 1 // all rules match at least one splitText
  /// 2 // fast match one and don't match other splitTexts
  /// 3 // strict mode, all splitTexts must have match.
  List<List<RegExpMatch>> splitAutoCompletingMatches(
      List<String> splitText, String stateName,
      {int matchMode = 3}) {
    assert(_runtimeContext != null);

    ///  一个 rule 对应 每个splitText在该rule中对应的 text 位置
    var matches = <List<RegExpMatch>>[];

    _runtimeContext![stateName]?.whereType<JParse>().forEach((jparse) {
      if (jparse.isConst && jparse is! GroupJParse) {
        matches.addAll(enumSplitMatches(splitText, jparse.constants!,
            matchMode: matchMode));
      } else if (jparse is GroupJParse &&
          jparse.groupDTokens.any((dtoken) => dtoken?.isEnum ?? false)) {
        // print(element.pattern);
        var constantRules = jparse.groupDTokens
            .map<List<String>>((dtoken) => _enumAllGiven(dtoken, stateName))
            .toList(growable: false);
        if (jparse.groupDTokens.any((dtoken) => dtoken is Templates)) {
          // List<List<Map<String, dynamic>>> message = [];
          var templateRes = <List<List<dynamic>>>[];

          //处理模板
          jparse.groupDTokens.forEach((dtoken) {
            var temp = <List<dynamic>>[];
            if (dtoken is Templates) {
              /// 处理变量
              var enumStrings = <int, Map<String, List<String>>>{};
              var dtokenVars = dtoken.vars;
              if (dtokenVars != null) {
                final enumvars = dtokenVars.enumvars;
                final enumnone = dtokenVars.enumnone;
                final patterns = dtokenVars.patterns;
                for (var i = 0; i < enumvars.length; i++) {
                  // step1: 优先根据 pattern 从 splittext 提取
                  if (patterns != null) {
                    var range =
                        splitTextRange(patterns[i] as String, splitText);
                    if (range.isNotEmpty) {
                      enumStrings[i] = {
                        'value': [range.join()],
                        'depart': range
                      };
                    }
                  }
                  // step2: 其次从枚举值中提取
                  if (enumStrings[i] == null) {
                    var enums = _enumAllGiven(enumvars[i], stateName);
                    if (enums.isNotEmpty) {
                      enumStrings[i] = {'value': enums, 'depart': []};
                    }
                  }
                  // step3: 最后从enumnone中提取兜底数据
                  if (enumStrings[i] == null && enumnone != null) {
                    enumStrings[i] = {
                      'value': [enumnone[i]],
                      'depart': []
                    };
                  }
                }
              }

              /// 替换变量
              dtoken.templates.forEach((template) {
                List<String> splits;

                /// 情况1： 没有变量直接添加
                if ((splits = template.split(r'\$'))
                    .every((item) => !item.contains(r'$'))) {
                  temp.add([splits.join(r'$'), []]);
                  return;
                }

                /// 情况2：有变量但没有定义变量则退出
                if (dtoken.vars == null) return;
                var splitExpands = <List<List<String>>>[
                  [splits, []]
                ];

                /// 情况3：有变量有定义
                /// 从 enumStrings 中去取
                /// ($/$1)对应第一个 $2对应第二个 $3对应第三个 依次类推..
                /// index   0           1           2 ..
                final varLength = dtokenVars?.enumvars.length ?? 0;
                for (var varIndex = varLength - 1; varIndex >= 0; varIndex--) {
                  var _splitExpands =
                      splitExpands.expand<List<List<String>>>((splitSL) {
                    var parts = splitSL[0];
                    var matched = splitSL[1];
                    if (parts.any((part) =>
                        part.contains(varName(varIndex, varLength)))) {
                      // 情况3.1：定义了变量但是 enumStrings没有 则去除该条规则
                      if (enumStrings[varIndex] == null) return [];
                      var res = <List<List<String>>>[];
                      enumStrings[varIndex]?['value']?.forEach((varValue) {
                        res.add([
                          parts
                              .map((part) => part.replaceAll(
                                  varName(varIndex, varLength), varValue))
                              .toList(),
                          [...matched, ...enumStrings[varIndex]!['depart']!]
                        ]);
                      });
                      // []..add([splitSL[0], splitSL[1]]);
                      return res;
                    } else {
                      return [splitSL];
                    }
                  }).toList();
                  splitExpands = _splitExpands;
                  temp.addAll(splitExpands
                      .map<List<dynamic>>((splitItem) =>
                          [splitItem[0].join(r'$'), splitItem[1]])
                      .toList());
                }
              });

              // return [[_inject(dtoken.templates[0],listVars), departVars]]
            }

            /// 占位
            templateRes.add(temp);
          });
          matches.addAll(enumSplitMatches(splitText, constantRules,
              matchMode: matchMode,
              templates: templateRes
                  .map((e) => e.map<String>((e) => e[0] as String).toList())
                  .toList(),
              matchedFromTemplate: templateRes
                  .map((e) => e
                      .map<List<String>>((e) =>
                          e[1].isEmpty ? <String>[] : e[1] as List<String>)
                      .toList())
                  .toList()));
        }
        matches.addAll(
            enumSplitMatches(splitText, constantRules, matchMode: matchMode));
      }
    });
    return matches;
  }

  /// Match wid to width
  /// errorText will be match
  List<String> autoCompleting(String errorText, String stateName) {
    assert(_runtimeContext != null);
    var matches = <Match>[];
    var warnedToUseSplitAutoCompleting = false;

    /// step1: find matches;
    _runtimeContext![stateName]?.whereType<JParse>().forEach((jparse) {
      if (jparse.isConst && jparse is! GroupJParse) {
        matches.addAll(enumAllMatches(errorText, jparse.constants!));
      } else if (jparse is GroupJParse &&
          jparse.groupDTokens.any((dtoken) => dtoken?.isEnum ?? false)) {
        // print(element.pattern);
        var constantRules = jparse.groupDTokens
            .map<List<String>>((dtoken) => _enumAllGiven(dtoken, stateName))
            .toList(growable: false);
        List<List<String>?>? templates;

        if (jparse.groupDTokens.any((dtoken) => dtoken is Templates)) {
          templates = jparse.groupDTokens.map((dtoken) {
            if (dtoken is Templates) {
              if (dtoken.vars != null) {
                if (warnedToUseSplitAutoCompleting == false) {
                  warnedToUseSplitAutoCompleting = true;
                }
                return [...dtoken.templates]..removeWhere((template) =>
                    template.replaceAll(r'\$', '').contains(r'$'));
              }
              return dtoken.templates;
            }
            return null;
          }).toList(growable: false);
        }
        matches.addAll(
            enumAllMatches(errorText, constantRules, templates: templates));
      }
    });

    /// step2: sort
    /// 从左到右 -> 优先
    /// start 小 ->
    /// input.length 小 ->
    matches.sort((left, right) {
      if (left.start < right.start) {
        return -1;
      } else if (left.input.length < right.input.length) {
        return -1;
      }
      return 1;
    });

    if (warnedToUseSplitAutoCompleting == true) {
      print(
          '\nWarning:Template Variable Injection only work in splitAutoCompleting!\n'
          r'All templates with symbol($) will be ignored');
    }

    /// step3: 用 Set 去重
    return (<String>{}..addAll(matches.map<String>((m) => m.input))).toList();
  }

  String pretty(String text,
      [List<String>? stack, Position? pos, Map<String, List<T>>? commondefs]) {
    return processedString(
        unprocessedTokens: getTokensUnprocessed(text, stack, pos, commondefs));
  }

  void configPrint() {
    print('''Settings:
+============================
|  $CONFIG_EVENT_DISPATCHER : $configEventDispatcher
|  $CONFIG_SAVING_RUNTIME_CONTEXT : $configSaveRuntimeContext
|  $CONFIG_STATE_WILL_LIST_TOKENS : $configListTokenStateNames
|  $CONFIG_USECACHE : $useCache
|  $CONFIG_ENUMSTRICT : $configEnumStrict
|  $CONFIG_GROUPCOUNT_STRICTCHECK : $configGroupCountStrictCheck
|  $CONFIG_MATCHLENGTH_STRICTCHECK : $configMatchLengthStrictCheck
|  $CONFIG_DEBUGPRINT_THROWERROR : $configDebugPrintThrowError
+============================
''');
    if (configListTokenStateNames == null || _tokensUnprocessed == null) return;
    print('''

Tokens by state:
=============================
''');
    var total = 0;
    configListTokenStateNames?.forEach((tokenObj) {
      var byToken = 0;
      print('''   ${tokenName(tokenObj)}:
    =========================
''');
      _tokensUnprocessed?.forEach((UnprocessedToken ut) {
        if (tokenObj == ut.token ||
            ut.token.toString() == 'Token.' + tokenName(tokenObj)) {
          byToken++;
          print('''   ${ut.stateName}:
    ${ut.match}
''');
        }
      });
      var byTokenPadding = 'total: $byToken'.padLeft(25);
      total += byToken;
      print('''
    =========================
    $byTokenPadding
''');
    });
    var totalPadding = 'total: $total'.padLeft(29);
    print('''
=============================
$totalPadding
''');
  }

  // Split ``text`` into (token type, text) pairs.
  //
  // ``stack`` is the initial stack (default: ``[root]``)
  //
  // The get_tokens_unprocessed() method must return an iterator or iterable
  // containing tuples in the form (index, token, value).
  // Stream<Tuple2<index, token, value>>
  List<UnprocessedToken>? _tokensUnprocessed;

  @override
  Iterable<UnprocessedToken> getTokensUnprocessed(String text,
          [List<String>? stack,
          Position? pos,
          Map<String, List<T>>? commondefs]) =>
      __innerGetTokensUnprocessed(text, stack, pos, commondefs);

  Iterable<UnprocessedToken> __innerGetTokensUnprocessed(String text,
      [List<String>? stack, Position? pos, Map<String, List<T>>? commondefs]) {
    // init 变量
    var ret = <UnprocessedToken>[];
    pos = pos ?? Position(0);

    /// merge common
    commondefs = _merge(commondefs, commonparses(commondefs));
    var parsedefs =

        /// merge all
        _expand(_merge(commondefs, parses));
    _runtimeContext = configSaveRuntimeContext ? parsedefs : null;
    final statestack = stack ?? List.from([root]);
    var statetokens = parsedefs[statestack.last]!;
    if (configDebuggable) {
      print('\n==================================');
      print('\n$this invoke getTokenUnprocessed');
      print('\nstatestack: $statestack');
      print('\nparsedefs: ${parsedefs.keys}');
      print('\n==================================');
    }

    var stateStartTrigger = false;
    var stateRestartTrigger = false;
    var stateEndTrigger = false;
    String? _willRestartFlag;
    // init 函数
    void _markStateAboutEventEmmitor(_) {
      stateStartTrigger = false;
      stateRestartTrigger = false;
      stateEndTrigger = false;
      _.forEach((T item) {
        if (item is EventEmmitor) {
          // ignore: unrelated_type_equality_checks
          if (item.dtoken == Token.EventOnStateWillStart) {
            stateStartTrigger = true;
          }
          // ignore: unrelated_type_equality_checks
          else if (item.dtoken == Token.EventOnStateWillRestart) {
            stateRestartTrigger = true;
          }
          // ignore: unrelated_type_equality_checks
          else if (item.dtoken == Token.EventOnStateWillEnd) {
            stateEndTrigger = true;
          }
        }
      });
    }

    String currentStateName() {
      return statestack.isEmpty ? root : statestack.last;
    }

    var _tempL = ret.length;
    Context _buildContext() {
      var lastL = _tempL;
      _tempL = ret.length;
      return TokenContext(parsedefs, ret, lastL);
    }

    bool _finishJumpState(newStates) {
      for (final state in newStates) {
        if (_doStackActionsNeedBreak(statestack, state,
                context: _buildContext()) ==
            true) {
          return true;
        }
      }
      statetokens = parsedefs[currentStateName()]!;
      // statestack.isEmpty ? parsedefs[root] : parsedefs[statestack.last];
      // _markStateAboutEventEmmitor(statetokens);
      // if (statetokens == null) debugger();
      return false;
    }

    while (true && pos.from < text.length) {
      var matched = false;
      var solidEventStateName = currentStateName();
      _markStateAboutEventEmmitor(statetokens);
      if (supportEventDispatching) {
        if (stateRestartTrigger && _willRestartFlag == currentStateName()) {
          configEventDispatcher!.dispatchEvent(
              Token.EventOnStateWillRestart,
              Token.EventOnStateWillRestart.toString().substring(6),
              solidEventStateName,
              _buildContext());
        } else if (stateStartTrigger) {
          configEventDispatcher!.dispatchEvent(
              Token.EventOnStateWillStart,
              Token.EventOnStateWillStart.toString().substring(6),
              solidEventStateName,
              _buildContext());
          _willRestartFlag = currentStateName();
        }
      }
      var _growableIterating = [...statetokens];
      while (_growableIterating.isNotEmpty) {
        final parse = _growableIterating.removeAt(0);
        // }
        // for (final parse in statetokens) {
        final pattern = parse.pattern;
        final token = parse is JParse ? parse.dtoken : parse.token;
        final newStates = parse.newStates;
        Token _tempt;
        if (token == (_tempt = Token.EventOnCondition) ||
            token == (_tempt = Token.EventOnConditionInclude) ||
            token == (_tempt = Token.EventOnRuleMissed) ||
            token == (_tempt = Token.EventOnRuleWillStart)) {
          if (supportEventDispatching) {
            // Token.EventOnCondition 具有跳转的功能
            if (configEventDispatcher!.dispatchEvent(
                    _tempt, pattern, solidEventStateName, _buildContext()) ==
                true) {
              if (_tempt == Token.EventOnCondition) {
                if (newStates != null && _finishJumpState(newStates)) {
                  return ret;
                }
              } else if (_tempt == Token.EventOnConditionInclude) {
                _growableIterating.insertAll(
                    0,
                    expandList(
                        (parse as JParsePackage).package.cast<T>().toList(),
                        parsedefs));
              }
            }
          }
        } else if (token == Token.EventOnStateWillEnd ||
            token == Token.EventOnStateWillRestart ||
            token == Token.EventOnStateWillStart) {
          continue;
        } else if (token == Token.IncludeOtherLexer) {
          var lexer = (parse is LexerJParse
              ? parse.lexer
              : (parse as LexerParse).lexer) as RegexLexer<T>;
          lexer.config = config;
          if (newStates != null) {
            for (final state in newStates) {
              _doStackActionsNeedBreak(statestack, state,
                  context: _buildContext(), onlyActions: true);
            }
          }
          statestack.add(lexer.root);
          var pos_before = pos.from;
          var res = lexer.__innerGetTokensUnprocessed(
              text, statestack, pos, commondefs);
          _runtimeContext = lexer._runtimeContext;
          ret.addAll(res);
          if (pos.from > pos_before) {
            matched = true;
            break;
          } else {
            continue;
          }
        }

        final regex = RegExp(
          pattern!,
          dotAll: flags.dotAll,
          unicode: flags.unicode,
          multiLine: flags.multiline,
          caseSensitive: flags.caseSensitive,
        );

        final m = regex.matchAsPrefix(text, pos.from);
        if (m != null) {
          if (token != null && m.group(0)!.isNotEmpty) {
            if (token == Token.ParseByGroups) {
              var groupTokens =
                  parse.groupTokens ?? (parse as GroupJParse).groupDTokens;
              if (m.groupCount != groupTokens.length &&
                  configGroupCountStrictCheck) {
                printThrow('''Warning: groupCount unequal
                r'$pattern' in $solidEventStateName gives ${m.groupCount} groups while tokens give ${groupTokens.length}
                Use null if known as blank match or give a Token.Text token to match the count.
                ''');
              }
              var willAdd = _bygroup(m, groupTokens, currentStateName());
              if (configMatchLengthStrictCheck) {
                var countLength = willAdd.fold<int>(
                    0,
                    (previousValue, element) =>
                        previousValue + element.match.length);
                var originLength = m.group(0)!.length;
                if (countLength != originLength) {
                  printThrow('''Warning: matchLength unequal
                r'$pattern' in $solidEventStateName matches ${m.group(0)} with length $originLength while tokens link as ${processedString(unprocessedTokens: willAdd)} with length $countLength
                Use null if known as blank match or give a Token.Text token to match the count.
                ''');
                }
              }
              ret.addAll(willAdd);
            } else {
              ret.add(UnprocessedToken(pos.from, token, m.group(0)!,
                  currentStateName(), debuggable));
            }
          }
          pos.from = m.end;
          if (newStates != null && _finishJumpState(newStates)) {
            return ret;
          }
          matched = true;
          break;
        }
      }
      if (supportEventDispatching && stateEndTrigger) {
        configEventDispatcher!.dispatchEvent(
            Token.EventOnStateWillEnd,
            Token.EventOnStateWillEnd.toString().substring(6),
            solidEventStateName,
            _buildContext());
      }
      if (!matched) {
        // We are here only if all state tokens have been considered
        // and there was not a match on any of them.
        // try {
        if (text[pos.from] == '\n') {
          _popTo(statestack, root);
          statetokens = parsedefs[root]!;
          ret.add(UnprocessedToken(pos.from, Token.Text, '\n',
              statestack.isEmpty ? root : statestack.last, debuggable));
          pos.from++;
          continue;
        }
        // print(
        //     'Error starts from "${text.substring(pos.from)}" in statestack: $statestack \n');
        ret.add(UnprocessedToken(pos.from, Token.Error, text[pos.from],
            currentStateName(), debuggable));
        pos.from++;
        // } on Exception catch (err) {
        //   print(err);
        //   break;
        // }
      }
    }
    return _tokensUnprocessed = ret;
  }

  bool _doStackActionsNeedBreak(List<String> statestack, String stateOrAction,
      {Context? context, bool onlyActions = false}) {
    if (stateOrAction == CLEAR) {
      statestack.clear();
    } else if (stateOrAction == BREAK) {
      return true;
    } else if (stateOrAction == POP) {
      _pop(statestack, 1);
    } else if (stateOrAction == POP2) {
      _pop(statestack, 2);
    } else if (stateOrAction == POPROOT) {
      _popTo(statestack, root);
    } else if (stateOrAction == PUSH) {
      statestack.add(statestack.isNotEmpty ? statestack.last : root);
    } else if (stateOrAction.startsWith('#event-match')) {
      if (supportEventDispatching) {
        configEventDispatcher!.dispatchEvent(
            Token.EventOnRuleMatchedEnterLeave,
            stateOrAction,
            statestack.isNotEmpty ? statestack.last : root,
            context);
      }
    } else if (stateOrAction.startsWith(r'#inside-')) {
    }

    /// add more dynamic way to control state stack.
    else if (stateOrAction.startsWith('#pop:')) {
      final popingState = stateOrAction.replaceFirst('#pop:', '');
      if (int.tryParse(popingState) != null) {
        _pop(statestack, int.parse(popingState));
      } else {
        _popTo(statestack, popingState);
      }

      /// end add by jackyanjiaqi.
    } else if (stateOrAction.startsWith('#on:')) {
      var rules1 = <List<dynamic>>[
        [
          r'#on:(.*)#pop:(.*)#do:\{(.*)\}#else:\{(.*)\}',
          [statestack, 1, 2, 3, 4]
        ],
        [
          r'#on:(.*)#pop:(.*)#do:\{(.*)\}',
          [statestack, 1, 2, 3, null]
        ],
        [
          r'#on:(.*)#pop:(.*)#else:\{(.*)\}',
          [statestack, 1, 2, null, 3]
        ],
        [
          r'#on:(.*)#pop:(.*)',
          [statestack, 1, 2, null, null]
        ],
        [
          r'#on:(.*)#do:\{(.*)\}#else:\{(.*)\}',
          [statestack, 1, null, 2, 3]
        ],
        [
          r'#on:(.*)#do:\{(.*)\}',
          [statestack, 1, null, 2, null]
        ],
        [
          r'#on:(.*)#else:\{(.*)\}',
          [statestack, 1, null, null, 2]
        ],
      ];

      /// Simple Rules But Need More Test
      // List<List<dynamic>> rules2 = [
      //   [
      //     r'#on:(.*)(#pop:(.*)|#do:\{(.*)\}|#else:\{(.*)\})+?',
      //     [statestack, 1, 3, 4, 5],
      //     {
      //       2: [3, 4, 5]
      //     }
      //   ]
      // ];
      return RegstrInvoke(stateOrAction, rules1, invoker: _onPopDoElse);

      /// skipping states when onlyAction == true;
    } else if (onlyActions == false) {
      statestack.add(stateOrAction);
      if (configDebuggable) print(statestack);
    }

    return false;

    /// end add by jackyanjiaqi.
  }

  // Callback that yields multiple actions for each group in the match.
  Iterable<UnprocessedToken> _bygroup(
      Match m, Iterable<Object?> tokens, String stateName) sync* {
    var groupIdx = 1;
    var pos = m.start;
    for (final token in tokens) {
      if (token == null) {
        groupIdx++;
        continue;
      }
      final s = m.group(groupIdx)!;
      if (token == Token.RecurseSameLexer) {
        yield* getTokensUnprocessed(s);
      } else {
        yield UnprocessedToken(pos, token, s, stateName, debuggable);
      }

      pos += s.length;
      groupIdx++;
    }
  }

  void _pop(List<String> statestack, int times) {
    while (statestack.isNotEmpty && times > 0) {
      statestack.removeLast();
      times--;
    }
    if (configDebuggable == true) print(statestack);
  }

  void _popTo(List<String> statestack, String target) {
    while (statestack.isNotEmpty && statestack.last != target) {
      statestack.removeLast();
    }
  }

  bool _onPopDoElse(List<String> statestack, String? onstate, String? popto,
      String? doAction, String? elseAction) {
    if (statestack.contains(onstate)) {
      if (popto != null) _popTo(statestack, popto);
      if (doAction != null) {
        return _doStackActionsNeedBreak(statestack, doAction);
      }
    } else if (elseAction != null) {
      return _doStackActionsNeedBreak(statestack, elseAction);
    }
    return false;
  }

  Map<String, List<T>> _merge(
      Map<String, List<T>>? merge1, Map<String, List<T>>? merge2) {
    final merged = <String, List<T>>{};
    if (merge1?.isNotEmpty ?? false) {
      for (final entry in merge1!.entries) {
        merged[entry.key] = entry.value;
      }
    }
    if (merge2?.isNotEmpty ?? false) {
      for (final entry in merge2!.entries) {
        merged[entry.key] = entry.value;
      }
    }
    return merged;
  }

  List<String>? _replaceOrAddNewStates(
      List<String>? replaced, List<String>? origin) {
    if (replaced != null && origin != null) {
      if (replaced.contains(INSIDE_MARK_ADDED)) {
        // 内层可加，可以递归累加
        if (origin.contains(INSIDE_MARK_ADDED)) {
          return [...replaced, ...origin];
          // 内层替换，只加到本层
        } else {
          return [...replaced, ...origin]
            ..removeWhere((element) => element == INSIDE_MARK_ADDED);
        }
      } else {
        return replaced;
      }
    } else {
      return replaced ?? origin;
    }
  }

  Iterable<T> expandList(List<T> parses, Map<String, List<T>> parsesMap,
      {List<String>? replaceNewState}) sync* {
    for (final p in parses) {
      if (p.token == Token.IncludeOtherParse) {
        yield* expandList(parsesMap[p.pattern]!, parsesMap,
            replaceNewState:
                _replaceOrAddNewStates(replaceNewState, p.newStates));
      } else if (replaceNewState != null) {
        /// Parse.include 批量替换 newState 避免污染原 Parse
        yield p.copy(
                newStates: _replaceOrAddNewStates(replaceNewState, p.newStates))
            as T;
      } else {
        yield p;
      }
    }
  }

  Map<String, List<T>> _expand(Map<String, List<T>> parsesMap) {
    final expanded = <String, List<T>>{};
    for (final entry in parsesMap.entries) {
      if (useCache) {
        // _cacheContext ??= {};
        if (_cacheContext.containsKey(entry.key)) {
          expanded[entry.key] = _cacheContext[entry.key]!;
        } else {
          expanded[entry.key] = expandList(entry.value, parsesMap).toList();
          _cacheContext[entry.key] = expanded[entry.key]!;
        }
      } else {
        expanded[entry.key] = expandList(entry.value, parsesMap).toList();
      }
    }
    return expanded;
  }
}
